@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -30,35 +30,36 @@
 #include "jam.h"
 #include "alloc.h"
 #include "thread.h"
+#include "arch.h"
 
 /* Trace GC heap mark/sweep phases - useful for debugging heap
  * corruption */
 #ifdef TRACEGC
-#define TRACE_GC(x) printf x
+#define TRACE_GC(x) printf x; fflush(stdout)
 #else
 #define TRACE_GC(x)
 #endif
 
 /* Trace class, object and array allocation */
 #ifdef TRACEALLOC
-#define TRACE_ALLOC(x) printf x
+#define TRACE_ALLOC(x) printf x; fflush(stdout)
 #else
 #define TRACE_ALLOC(x)
 #endif
 
 /* Trace object finalization */
 #ifdef TRACEFNLZ
-#define TRACE_FNLZ(x) printf x
+#define TRACE_FNLZ(x) printf x; fflush(stdout)
 #else
 #define TRACE_FNLZ(x)
 #endif
 
-#define OBJECT_GRAIN		8
-#define	ALLOC_BIT		1
+#define OBJECT_GRAIN            8
+#define ALLOC_BIT               1
 
-#define HEADER(ptr)		*((unsigned int*)ptr)
-#define HDR_SIZE(hdr)		(hdr & ~(ALLOC_BIT|FLC_BIT))
-#define HDR_ALLOCED(hdr)	(hdr & ALLOC_BIT)
+#define HEADER(ptr)             *((unsigned int*)ptr)
+#define HDR_SIZE(hdr)           (hdr & ~(ALLOC_BIT|FLC_BIT))
+#define HDR_ALLOCED(hdr)        (hdr & ALLOC_BIT)
 
 /* 1 word header format
   31                                       210
@@ -103,18 +104,18 @@ static VMWaitLock run_fnlzr_lock;
 
 static Object *oom;
 
-#define LIST_INCREMENT		1000
+#define LIST_INCREMENT          1000
 
-#define LOG_BYTESPERBIT		LOG_OBJECT_GRAIN /* 1 mark bit for every OBJECT_GRAIN bytes of heap */
-#define LOG_MARKSIZEBITS	5
-#define MARKSIZEBITS		32
+#define LOG_BYTESPERBIT         LOG_OBJECT_GRAIN /* 1 mark bit for every OBJECT_GRAIN bytes of heap */
+#define LOG_MARKSIZEBITS        5
+#define MARKSIZEBITS            32
 
-#define MARKENTRY(ptr)	((((char*)ptr)-heapbase)>>(LOG_BYTESPERBIT+LOG_MARKSIZEBITS))
-#define MARKOFFSET(ptr)	(((((char*)ptr)-heapbase)>>LOG_BYTESPERBIT)&(MARKSIZEBITS-1))
-#define MARK(ptr)	markBits[MARKENTRY(ptr)]|=1<<MARKOFFSET(ptr)
-#define IS_MARKED(ptr)	(markBits[MARKENTRY(ptr)]&(1<<MARKOFFSET(ptr)))
+#define MARKENTRY(ptr)  ((((char*)ptr)-heapbase)>>(LOG_BYTESPERBIT+LOG_MARKSIZEBITS))
+#define MARKOFFSET(ptr) (((((char*)ptr)-heapbase)>>LOG_BYTESPERBIT)&(MARKSIZEBITS-1))
+#define MARK(ptr)       markBits[MARKENTRY(ptr)]|=1<<MARKOFFSET(ptr)
+#define IS_MARKED(ptr)  (markBits[MARKENTRY(ptr)]&(1<<MARKOFFSET(ptr)))
 
-#define IS_OBJECT(ptr)	(((char*)ptr) > heapbase) && \
+#define IS_OBJECT(ptr)  (((char*)ptr) > heapbase) && \
                         (((char*)ptr) < heaplimit) && \
                         !(((unsigned int)ptr)&(OBJECT_GRAIN-1))
 
@@ -139,11 +140,11 @@ void initialiseAlloc(int min, int max, int verbose) {
     min = max;
     if(mem == NULL) {
 #else
-    char *mem = (char*)mmap(0, max, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
+    char *mem = (char*)mmap(0, max, PROT_READ|PROT_WRITE, MAP_PRIVATE|MAP_ANON, -1, 0);
     if(mem == MAP_FAILED) {
 #endif
         perror("Aborting the VM -- couldn't allocate the heap");
-	exitVM(1);
+        exitVM(1);
     }
 
     /* Align heapbase so that start of heap + HEADER_SIZE is object aligned */
@@ -250,9 +251,9 @@ static void doMark(Thread *self) {
                 run_finaliser_end = run_finaliser_size;
                 run_finaliser_size += LIST_INCREMENT;
                 run_finaliser_list = (Object**)sysRealloc(run_finaliser_list,
-				run_finaliser_size*sizeof(Object*));
+                                run_finaliser_size*sizeof(Object*));
             }
-	    run_finaliser_end = run_finaliser_end%run_finaliser_size;
+            run_finaliser_end = run_finaliser_end%run_finaliser_size;
             run_finaliser_list[run_finaliser_end++] = ob;
         } else {
             has_finaliser_list[j++] = ob;
@@ -264,11 +265,11 @@ static void doMark(Thread *self) {
     if(j != has_finaliser_count) {
         has_finaliser_count = j;
 
-	/* Extra finalizers to be ran, so signal the finalizer thread
-	   in case it needs waking up.  It won't run until it's
-	   resumed */
+        /* Extra finalizers to be ran, so signal the finalizer thread
+           in case it needs waking up.  It won't run until it's
+           resumed */
 
-	notifyVMWaitLock(run_fnlzr_lock, self);
+        notifyVMWaitLock(run_fnlzr_lock, self);
     }
     unlockVMWaitLock(run_fnlzr_lock, self);
 }
@@ -306,7 +307,7 @@ static int doSweep(Thread *self) {
             unmarked++;
 
             TRACE_GC(("FREE: Freeing ob @ 0x%x class %s - start of block\n", ob,
-				    ob->class ? CLASS_CB(ob->class)->name : "?"));
+                                    ob->class ? CLASS_CB(ob->class)->name : "?"));
         }
         else
             TRACE_GC(("FREE: Unalloced block @ 0x%x size %d - start of block\n", ptr, size));
@@ -450,14 +451,14 @@ static int runFinalizers() {
         ob = run_finaliser_list[run_finaliser_start];
 
         unlockVMWaitLock(run_fnlzr_lock, self);
-	enableSuspend(self);
+        enableSuspend(self);
 
-	/* Run the finalizer method */
+        /* Run the finalizer method */
         executeMethod(ob, CLASS_CB(ob->class)->finalizer);
 
-	/* We entered with suspension off - nothing
-	 * else interesting on stack, so use previous
-	 * stack top. */
+        /* We entered with suspension off - nothing
+         * else interesting on stack, so use previous
+         * stack top. */
 
         disableSuspend0(self, getStackTop(self));
         lockVMWaitLock(run_fnlzr_lock, self);
@@ -465,7 +466,7 @@ static int runFinalizers() {
         /* Clear any exceptions - exceptions thrown in finalizers are
            silently ignored */
 
-	clearException();
+        clearException();
     } while(++run_finaliser_start != run_finaliser_end);
 
     run_finaliser_start = run_finaliser_size;
@@ -614,17 +615,17 @@ void *gcMalloc(int len) {
 #endif
         }
 
-	if(verbosegc)
+        if(verbosegc)
             printf("<GC: Alloc attempt for %d bytes failed (state %d).>\n", n, state);
 
-	switch(state) {
+        switch(state) {
 
             case 0:
                 largest = gc0();
                 if(n <= largest)
-		    break;
+                    break;
 
-	        state = 1;
+                state = 1;
 
             case 1: {
                 int res;
@@ -633,50 +634,50 @@ void *gcMalloc(int len) {
                 res = runFinalizers();
                 lockVMLock(heap_lock, self);
 
-		if(state == 1) {
+                if(state == 1) {
                     if(res) {
                         largest = gc0();
                         if(n <= largest) {
                             state = 0;
-			    break;
-		        }
+                            break;
+                        }
                     }
-		    state = 2;
-		}
-		break;
+                    state = 2;
+                }
+                break;
 
             case 2:
                 if(heaplimit < heapmax) {
-	            expandHeap(n);
-		    state = 0;
-		    break;
+                    expandHeap(n);
+                    state = 0;
+                    break;
                 } else {
                     if(verbosegc)
                         printf("<GC: Stack at maximum already - completely out of heap space>\n");
 
-		    state = 3;
+                    state = 3;
                     /* Reset next allocation block
-		     * to beginning of list */
+                     * to beginning of list */
                     chunkpp = &freelist;
                     enableSuspend(self);
                     unlockVMLock(heap_lock, self);
                     signalException("java/lang/OutOfMemoryError", NULL);
                     return NULL;
                 }
-		break;
+                break;
             }
 
             case 3:
-		/* Already throwing an OutOfMemoryError in some thread.  In both
-		 * cases, throw an already prepared OOM (no stacktrace).  Could have a
-		 * per-thread flag, so we try to throw a new OOM in each thread, but
-		 * if we're this low on memory I doubt it'll make much difference.
-		 */
+                /* Already throwing an OutOfMemoryError in some thread.  In both
+                 * cases, throw an already prepared OOM (no stacktrace).  Could have a
+                 * per-thread flag, so we try to throw a new OOM in each thread, but
+                 * if we're this low on memory I doubt it'll make much difference.
+                 */
 
-		state = 0;
+                state = 0;
                 enableSuspend(self);
                 unlockVMLock(heap_lock, self);
-		setException(oom);
+                setException(oom);
                 return NULL;
                 break;
         }
@@ -706,15 +707,15 @@ gotIt:
 }
 
 /* Object roots :-
-	classes
-	class statics
-	interned strings
-	thread Java stacks
+        classes
+        class statics
+        interned strings
+        thread Java stacks
         thread C stacks
-	thread registers (put onto C stack)
+        thread registers (put onto C stack)
         JNI refs
-	    - locals (scanned as part of JavaStack)
-	    - globals
+            - locals (scanned as part of JavaStack)
+            - globals
 */
 
 void markClassStatics(Class *class) {
@@ -867,8 +868,8 @@ int maxHeapMem() {
 void asyncGCThreadLoop(Thread *self) {
     for(;;) {
         threadSleep(self, 1000, 0);
-	if(systemIdle(self))
-	    gc1();
+        if(systemIdle(self))
+            gc1();
     }
 }
 
@@ -881,7 +882,7 @@ void finalizerThreadLoop(Thread *self) {
 
     for(;;) {
         lockVMWaitLock(run_fnlzr_lock, self);
-	waitVMWaitLock(run_fnlzr_lock, self);
+        waitVMWaitLock(run_fnlzr_lock, self);
         unlockVMWaitLock(run_fnlzr_lock, self);
         runFinalizers();
     }
@@ -895,7 +896,7 @@ void initialiseGC(int noasyncgc) {
     Class *oom_clazz = findSystemClass("java/lang/OutOfMemoryError");
     if(exceptionOccured()) {
         printException();
-	exitVM(1);
+        exitVM(1);
 
     }
 
@@ -918,11 +919,11 @@ void initialiseGC(int noasyncgc) {
     disableSuspend(self = threadSelf());                                           \
     lockVMLock(has_fnlzr_lock, self);                                              \
     TRACE_FNLZ(("Object @0x%x type %s has a finalize method...\n",                 \
-			                          ob, CLASS_CB(ob->class)->name)); \
+                                                  ob, CLASS_CB(ob->class)->name)); \
     if(has_finaliser_count == has_finaliser_size) {                                \
         has_finaliser_size += LIST_INCREMENT;                                      \
         has_finaliser_list = (Object**)sysRealloc(has_finaliser_list,              \
-			                       has_finaliser_size*sizeof(Object*));\
+                                               has_finaliser_size*sizeof(Object*));\
     }                                                                              \
                                                                                    \
     has_finaliser_list[has_finaliser_count++] = ob;                                \
@@ -1034,7 +1035,7 @@ Object *allocMultiArray(Class *array_class, int dim, int *count) {
         Class *aclass = findArrayClassFromClass(CLASS_CB(array_class)->name+1, array_class);
         array = allocArray(array_class, *count, 4);
 
-	if(array == NULL)
+        if(array == NULL)
             return NULL;
 
         for(i = 1; i <= *count; i++)
@@ -1084,7 +1085,8 @@ Object *cloneObject(Object *ob) {
     if(clone != NULL) {
         memcpy(clone, ob, size);
 
-	clone->lock = 0;
+        clone->lock = 0;
+        MBARRIER();
 
         if(CLASS_CB(clone->class)->finalizer != NULL)
             ADD_FINALIZED_OBJECT(clone);
