@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -22,7 +22,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
-
 #include <sys/stat.h>
 
 #include "jam.h"
@@ -30,9 +29,8 @@
 #include "thread.h"
 #include "lock.h"
 #include "hash.h"
-#include "class_md.h"
-
 #include "zip.h"
+#include "arch.h"
 
 #define PREPARE(ptr) ptr
 #define SCAVENGE(ptr) FALSE
@@ -70,18 +68,18 @@ static HashTable loaded_classes;
 static Class *prim_classes[MAX_PRIM_CLASSES];
 
 /* Macros for reading data values from class files - values
-   are in big endian format, and non-aligned.  See class_md.h
+   are in big endian format, and non-aligned.  See arch.h
    for READ_DBL - this is platform dependent */
 
-#define READ_U1(v,p,l)	v = *p++
+#define READ_U1(v,p,l)  v = *p++
 #define READ_U2(v,p,l)  v = (p[0]<<8)|p[1]; p+=2
-#define READ_U4(v,p,l)	v = (p[0]<<24)|(p[1]<<16)|(p[2]<<8)|p[3]; p+=4
-#define READ_U8(v,p,l)	v = ((u8)p[0]<<56)|((u8)p[1]<<48)|((u8)p[2]<<40) \
+#define READ_U4(v,p,l)  v = (p[0]<<24)|(p[1]<<16)|(p[2]<<8)|p[3]; p+=4
+#define READ_U8(v,p,l)  v = ((u8)p[0]<<56)|((u8)p[1]<<48)|((u8)p[2]<<40) \
                             |((u8)p[3]<<32)|((u8)p[4]<<24)|((u8)p[5]<<16) \
                             |((u8)p[6]<<8)|(u8)p[7]; p+=8
                             
-#define	READ_INDEX(v,p,l)		READ_U2(v,p,l)
-#define READ_TYPE_INDEX(v,cp,t,p,l)	READ_U2(v,p,l)
+#define READ_INDEX(v,p,l)               READ_U2(v,p,l)
+#define READ_TYPE_INDEX(v,cp,t,p,l)     READ_U2(v,p,l)
 
 
 static Class *addClassToHash(Class *class) {
@@ -101,15 +99,37 @@ static Class *addClassToHash(Class *class) {
 static void prepareClass(Class *class) {
     ClassBlock *cb = CLASS_CB(class);
 
-    if(strcmp(cb->name,"java/lang/Class") == 0)
-       class->class = class;
+    if(strcmp(cb->name, "java/lang/Class") == 0)
+       java_lang_Class = class->class = class;
     else {
        if(java_lang_Class == NULL)
-          java_lang_Class = findSystemClass0("java/lang/Class");
+          findSystemClass0("java/lang/Class");
        class->class = java_lang_Class;
     }
 }
 
+static void freeClassData(Class *class) {
+    ClassBlock *cb = CLASS_CB(class);
+    int i;
+
+    free((void*)cb->constant_pool.type);
+    free(cb->constant_pool.info);
+    free(cb->interfaces);
+    free(cb->fields);
+
+    for(i = 0; i < cb->methods_count; i++) {
+        MethodBlock *mb = &cb->methods[i];
+
+        free(mb->code);
+        free(mb->exception_table);
+        free(mb->line_no_table);
+        free(mb->throw_table);
+    }
+
+    free(cb->methods);
+    free(cb->inner_classes);
+}
+
 Class *defineClass(char *classname, char *data, int offset, int len, Object *class_loader) {
     unsigned char *ptr = (unsigned char *)data+offset;
     int cp_count, intf_count, i;
@@ -150,15 +170,15 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
         CP_TYPE(constant_pool,i) = tag;
 
         switch(tag) {
-	   case CONSTANT_Class:
-	   case CONSTANT_String:
+           case CONSTANT_Class:
+           case CONSTANT_String:
                READ_INDEX(CP_INFO(constant_pool,i), ptr, len);
                break;
 
-	   case CONSTANT_Fieldref:
-	   case CONSTANT_Methodref:
-	   case CONSTANT_NameAndType:
-	   case CONSTANT_InterfaceMethodref:
+           case CONSTANT_Fieldref:
+           case CONSTANT_Methodref:
+           case CONSTANT_NameAndType:
+           case CONSTANT_InterfaceMethodref:
            {
                u2 idx1, idx2;
 
@@ -168,22 +188,22 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
                break;
            }
 
-	   case CONSTANT_Integer:
-	   case CONSTANT_Float:
+           case CONSTANT_Integer:
+           case CONSTANT_Float:
                READ_U4(CP_INFO(constant_pool,i), ptr, len);
                break;
 
-	   case CONSTANT_Long:
+           case CONSTANT_Long:
                READ_U8(*(u8 *)&(CP_INFO(constant_pool,i)), ptr, len);
                i++;
                break;
                
-	   case CONSTANT_Double:
+           case CONSTANT_Double:
                READ_DBL(*(u8 *)&(CP_INFO(constant_pool,i)), ptr, len);
                i++;
                break;
 
-	   case CONSTANT_Utf8:
+           case CONSTANT_Utf8:
            {
                int length;
                unsigned char *buff, *utf8;
@@ -203,7 +223,7 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
                break;
            }
 
-	   default:
+           default:
                signalException("java/lang/ClassFormatError", "bad constant pool tag");
                return NULL;
         }
@@ -219,6 +239,8 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
         return NULL;
     }
 
+    prepareClass(class);
+
     if(strcmp(classblock->name, "java/lang/Object") == 0) {
         READ_U2(super_idx, ptr, len);
         if(super_idx) {
@@ -318,6 +340,8 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
               memcpy(method->code, ptr, code_length);
               ptr += code_length;
 
+              method->code_size = code_length;
+
               READ_U2(method->exception_table_size, ptr, len);
               method->exception_table = (ExceptionTableEntry *)
                   sysMalloc(method->exception_table_size*sizeof(ExceptionTableEntry));
@@ -345,13 +369,13 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
                      method->line_no_table = (LineNoTableEntry *)
                          sysMalloc(method->line_no_table_size*sizeof(LineNoTableEntry));
 
-		     for(j = 0; j < method->line_no_table_size; j++) {
+                     for(j = 0; j < method->line_no_table_size; j++) {
                          LineNoTableEntry *entry = &method->line_no_table[j];              
-			 
-			 READ_U2(entry->start_pc, ptr, len);
-			 READ_U2(entry->line_no, ptr, len);
-		     }
-		 } else
+                         
+                         READ_U2(entry->start_pc, ptr, len);
+                         READ_U2(entry->line_no, ptr, len);
+                     }
+                 } else
                      ptr += attr_length;
               }
            } else
@@ -381,7 +405,7 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
        if(strcmp(attr_name, "SourceFile") == 0) {
            u2 file_name_idx;
            READ_U2(file_name_idx, ptr, len);
-	   classblock->source_file_name = CP_UTF8(constant_pool, file_name_idx);
+           classblock->source_file_name = CP_UTF8(constant_pool, file_name_idx);
        } else
            if(strcmp(attr_name, "InnerClasses") == 0) {
                int j, size;
@@ -411,17 +435,18 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
                ptr += attr_length;
     }
 
-    classblock->flags = CLASS_LOADED;
-
-    if((found = addClassToHash(class)) != class)
-        return found;
-
     classblock->super = super_idx ? resolveClass(class, super_idx, FALSE) : NULL;
 
     if(exceptionOccured())
        return NULL;
 
-    prepareClass(class);
+    classblock->flags = CLASS_LOADED;
+
+    if((found = addClassToHash(class)) != class) {
+        freeClassData(class);
+        return found;
+    }
+
     return class;
 }
 
@@ -471,7 +496,8 @@ createArrayClass(char *classname, Object *class_loader) {
     classblock->class_loader = elem_cb->class_loader;
 
     /* The array's visibility (i.e. public, etc.) is that of the element */
-    classblock->access_flags = (elem_cb->access_flags & ~ACC_INTERFACE) | ACC_FINAL | ACC_ABSTRACT;
+    classblock->access_flags = (elem_cb->access_flags & ~ACC_INTERFACE) |
+                               ACC_FINAL | ACC_ABSTRACT;
     classblock->dim = dim;
 
     prepareClass(class);
@@ -510,24 +536,25 @@ createPrimClass(char *classname, int index) {
 
 #define MRNDA_CACHE_SZE 10
 
-#define resizeMTable(method_table, method_table_size, miranda, count)                \
-{                                                                                    \
-    method_table = (MethodBlock**)sysRealloc(method_table,                           \
-                  (method_table_size + count) * sizeof(MethodBlock*));               \
-                                                                                     \
-    memcpy(&method_table[method_table_size], miranda, count * sizeof(MethodBlock*)); \
-    method_table_size += count;                                                      \
+#define resizeMTable(method_table, method_table_size, miranda, count)  \
+{                                                                      \
+    method_table = (MethodBlock**)sysRealloc(method_table,             \
+                  (method_table_size + count) * sizeof(MethodBlock*)); \
+                                                                       \
+    memcpy(&method_table[method_table_size], miranda,                  \
+                               count * sizeof(MethodBlock*));          \
+    method_table_size += count;                                        \
 }
 
-#define fillinMTable(method_table, methods, methods_count)                           \
-{                                                                                    \
-    int i;                                                                           \
-    for(i = 0; i < methods_count; i++, methods++) {                                  \
-        if((methods->access_flags & ACC_STATIC) ||                                   \
-               (methods->access_flags & ACC_PRIVATE) || (methods->name[0] == '<'))   \
-            continue;                                                                \
-        method_table[methods->method_table_index] = methods;                         \
-    }                                                                                \
+#define fillinMTable(method_table, methods, methods_count)             \
+{                                                                      \
+    int i;                                                             \
+    for(i = 0; i < methods_count; i++, methods++) {                    \
+        if((methods->access_flags & (ACC_STATIC | ACC_PRIVATE)) ||     \
+               (methods->name[0] == '<'))                              \
+            continue;                                                  \
+        method_table[methods->method_table_index] = methods;           \
+    }                                                                  \
 }
 
 void linkClass(Class *class) {
@@ -537,7 +564,6 @@ void linkClass(Class *class) {
    int spr_mthd_tbl_sze = 0;
    MethodBlock **method_table = NULL;
    MethodBlock **spr_mthd_tbl;
-   MethodBlock *finalizer = NULL;
    int new_methods_count = 0;
    int spr_imthd_tbl_sze = 0;
    int method_table_size;
@@ -546,6 +572,8 @@ void linkClass(Class *class) {
    int field_offset = 0;
    int itbl_offset_count = 0;
    int itbl_idx, i, j;
+   MethodBlock *finalizer;
+   static MethodBlock *obj_fnlzr_mthd = NULL;
 
    Class *super = (cb->access_flags & ACC_INTERFACE) ? NULL : cb->super;
 
@@ -570,7 +598,6 @@ void linkClass(Class *class) {
       spr_imthd_tbl = super_cb->imethod_table;
       spr_mthd_tbl_sze = super_cb->method_table_size;
       spr_imthd_tbl_sze = super_cb->imethod_table_size;
-      finalizer = super_cb->finalizer;
    }
 
    /* prepare fields */
@@ -626,19 +653,20 @@ void linkClass(Class *class) {
            mb->max_locals = mb->args_count;
            mb->max_stack = 0;
        }
+#ifdef DIRECT
+       else  {
+           /* Set the bottom bit of the pointer to indicate the
+              method is unprepared */
+           mb->code = ((char*)mb->code) + 1;
+       }
+#endif
 
        /* Static, private or init methods aren't dynamically invoked, so
-	 don't stick them in the table to save space */
+         don't stick them in the table to save space */
 
-       if((mb->access_flags & ACC_STATIC) || (mb->access_flags & ACC_PRIVATE) ||
-	       (mb->name[0] == '<'))
+       if((mb->access_flags & (ACC_STATIC | ACC_PRIVATE)) || (mb->name[0] == '<'))
            continue;
 
-       /* check for finalizer */
-
-       if(super && strcmp(mb->name, "finalize") == 0 && strcmp(mb->type, "()V") == 0)
-	   finalizer = mb;
-
        /* if it's overriding an inherited method, replace in method table */
 
        for(j = 0; j < spr_mthd_tbl_sze; j++)
@@ -652,8 +680,6 @@ void linkClass(Class *class) {
            mb->method_table_index = spr_mthd_tbl_sze + new_methods_count++;
    }
 
-   cb->finalizer = finalizer;
-
    /* construct method table */
 
    method_table_size = spr_mthd_tbl_sze + new_methods_count;
@@ -661,10 +687,11 @@ void linkClass(Class *class) {
    if(!(cb->access_flags & ACC_INTERFACE)) {
        method_table = (MethodBlock**)sysMalloc(method_table_size * sizeof(MethodBlock*));
 
+       /* Copy parents method table to the start */
        memcpy(method_table, spr_mthd_tbl, spr_mthd_tbl_sze * sizeof(MethodBlock*));
 
-       /* fill in method table */
-
+       /* fill in the additional methods -- we use a
+          temporary because fillinMtable alters mb */
        mb = cb->methods;
        fillinMTable(method_table, mb, cb->methods_count);
    }
@@ -700,7 +727,7 @@ void linkClass(Class *class) {
        for(j = 0; j < intf_cb->imethod_table_size; j++) {
            Class *spr_intf = intf_cb->imethod_table[j].interface;
            cb->imethod_table[itbl_idx++].interface = spr_intf;
-	   itbl_offset_count += CLASS_CB(spr_intf)->method_table_size;
+           itbl_offset_count += CLASS_CB(spr_intf)->method_table_size;
        }
    }
 
@@ -724,27 +751,27 @@ void linkClass(Class *class) {
            for(j = 0; j < intf_cb->methods_count; j++) {
                MethodBlock *intf_mb = &intf_cb->methods[j];
 
-               if((intf_mb->access_flags & ACC_STATIC) ||
-		      (intf_mb->access_flags & ACC_PRIVATE) || (intf_mb->name[0] == '<'))
+               if((intf_mb->access_flags & (ACC_STATIC | ACC_PRIVATE)) ||
+                      (intf_mb->name[0] == '<'))
                    continue;
 
                for(mtbl_idx = 0; mtbl_idx < method_table_size; mtbl_idx++)
                    if(strcmp(intf_mb->name, method_table[mtbl_idx]->name) == 0 &&
-		           strcmp(intf_mb->type, method_table[mtbl_idx]->type) == 0) {
+                           strcmp(intf_mb->type, method_table[mtbl_idx]->type) == 0) {
                        *offsets_pntr++ = mtbl_idx;
-		       break;
+                       break;
                    }
 
-	       if(mtbl_idx == method_table_size) {
+               if(mtbl_idx == method_table_size) {
 
-	       /* didn't find it - add a dummy abstract method (a so-called
-                  miranda method).  If it's invoked we'll get an abstract
-                  method error */
+                   /* didn't find it - add a dummy abstract method (a so-called
+                      miranda method).  If it's invoked we'll get an abstract
+                      method error */
 
                    int k;
                    for(k = 0; k < miranda_count; k++)
                        if(strcmp(intf_mb->name, miranda[k]->name) == 0 &&
-	                           strcmp(intf_mb->type, miranda[k]->type) == 0)
+                                   strcmp(intf_mb->type, miranda[k]->type) == 0)
                            break;
                            
                    *offsets_pntr++ = method_table_size + k;
@@ -771,9 +798,9 @@ void linkClass(Class *class) {
                    (cb->methods_count + num_mirandas) * sizeof(MethodBlock));
 
            /* If the realloc of the method list gave us a new pointer, the pointers
-              to them in the method table are now wrong.  Note, fillInMTable will
-              leave mb pointing to the end */
+              to them in the method table are now wrong. */
            if(mb != cb->methods) {
+               /*  mb will be left pointing to the end of the methods */
                cb->methods = mb;
                fillinMTable(method_table, mb, cb->methods_count);
            } else
@@ -797,6 +824,27 @@ void linkClass(Class *class) {
    cb->method_table = method_table;
    cb->method_table_size = method_table_size;
 
+   /* handle finalizer */
+
+   /* If this is Object find the finalize method.  All subclasses will
+      have it in the same place in the method table.  Note, Object
+      should always have a valid finalizer -- but check just in case */
+
+   if(cb->super == NULL) {
+       finalizer = findMethod(class, "finalize", "()V");
+       if(finalizer && !(finalizer->access_flags & (ACC_STATIC | ACC_PRIVATE)))
+           obj_fnlzr_mthd = finalizer;
+   }
+
+   /* Store the finalizer only if it's overridden Object's.  We don't
+      want to finalize every object, and Object's imp is empty */
+
+   if(super && obj_fnlzr_mthd && (finalizer =
+               method_table[obj_fnlzr_mthd->method_table_index]) != obj_fnlzr_mthd)
+       cb->finalizer = finalizer;
+   else
+       cb->finalizer = NULL;
+
    cb->flags = CLASS_LINKED;
 
 unlock:
@@ -1079,6 +1127,7 @@ Class *findNonArrayClassFromClassLoader(char *classname, Object *loader) {
     if(class == NULL) {
         char *dot_name = slash2dots(classname);
         Object *string = createString(dot_name);
+        Object *excep;
 
         free(dot_name);
         if(string == NULL)
@@ -1096,9 +1145,9 @@ Class *findNonArrayClassFromClassLoader(char *classname, Object *loader) {
         class = *(Class**)executeMethod(loader,
                     CLASS_CB(loader->class)->method_table[loadClass_mtbl_idx], string);
 
-        if(exceptionOccured()) {
+        if((excep = exceptionOccured())) {
             clearException();
-            signalException("java/lang/NoClassDefFoundError", classname);
+            signalChainedException("java/lang/NoClassDefFoundError", classname, excep);
             return NULL;
         }
 
@@ -1182,11 +1231,11 @@ int parseBootClassPath(char *cp_var) {
 
     for(i = 0, start = pntr = cp; *pntr; pntr++) {
         if(*pntr == ':') {
-	    if(start != pntr) {
+            if(start != pntr) {
                 *pntr = '\0';
-	        i++;
+                i++;
             }
-	    start = pntr+1;
+            start = pntr+1;
         }
     }
     if(start != pntr)
@@ -1230,7 +1279,7 @@ char *getClassPath() {
 
 char *setBootClassPath(char *cmdlne_bcp, char bootpathopt) {
 #ifdef USE_ZIP
-    char *dflt_bcp = INSTALL_DIR"/share/jamvm/classes:"CLASSPATH_INSTALL_DIR"/share/classpath/glibj.zip";
+    char *dflt_bcp = INSTALL_DIR"/share/jamvm/classes.zip:"CLASSPATH_INSTALL_DIR"/share/classpath/glibj.zip";
 #else
     char *dflt_bcp = INSTALL_DIR"/share/jamvm/classes:"CLASSPATH_INSTALL_DIR"/share/classpath";
 #endif
@@ -1244,8 +1293,10 @@ char *setBootClassPath(char *cmdlne_bcp, char bootpathopt) {
                 strcat(strcat(strcpy(bootpath, cmdlne_bcp), ":"), dflt_bcp);
         } else
             bootpath = cmdlne_bcp;
-    } else
-        bootpath = dflt_bcp;
+    } else {
+        char *env = getenv("BOOTCLASSPATH");
+        bootpath = env ? env : dflt_bcp;
+    }
 
     return bootpath;
 }
@@ -1290,7 +1341,7 @@ void initialiseClass(char *classpath, char *bootpath, char bootpathopt, int verb
 
     if(!(bcp && parseBootClassPath(bcp))) {
         printf("bootclasspath is empty!\n");
-	exitVM(1);
+        exitVM(1);
     }
 
     verbose = verboseclass;
