@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -55,7 +55,7 @@ Object *exceptionOccured() {
    return getExecEnv()->exception; 
 }
 
-void signalException(char *excep_name, char *message) {
+void signalChainedException(char *excep_name, char *message, Object *cause) {
     if(VM_initing) {
         fprintf(stderr, "Exception occurred while VM initialising.\n");
         if(message)
@@ -71,8 +71,16 @@ void signalException(char *excep_name, char *message) {
             Object *str = message == NULL ? NULL : Cstr2String(message);
             MethodBlock *init = lookupMethod(exception,
                                           "<init>", "(Ljava/lang/String;)V");
-	    if(exp && init) {
+            if(exp && init) {
                 executeMethod(exp, init, str);
+
+                if(cause && !exceptionOccured()) {
+                    MethodBlock *mb = lookupMethod(exception, "initCause",
+                                            "(Ljava/lang/Throwable;)Ljava/lang/Throwable;");
+                    if(mb)
+                        executeMethod(exp, mb, cause);
+                }
+
                 getExecEnv()->exception = exp;
             }
         }
@@ -108,15 +116,15 @@ void printException() {
         if(ee->exception) {
             fprintf(stderr, "Exception occured while printing exception (%s)...\n",
                             CLASS_CB(ee->exception->class)->name);
-	    fprintf(stderr, "Original exception was %s\n", CLASS_CB(exception->class)->name);
+            fprintf(stderr, "Original exception was %s\n", CLASS_CB(exception->class)->name);
         }
     }
 }
 
-unsigned char *findCatchBlockInMethod(MethodBlock *mb, Class *exception, unsigned char *pc_pntr) {
+CodePntr findCatchBlockInMethod(MethodBlock *mb, Class *exception, CodePntr pc_pntr) {
     ExceptionTableEntry *table = mb->exception_table;
     int size = mb->exception_table_size;
-    int pc = pc_pntr - mb->code;
+    int pc = pc_pntr - ((CodePntr)mb->code);
     int i;
  
     for(i = 0; i < size; i++)
@@ -135,22 +143,22 @@ unsigned char *findCatchBlockInMethod(MethodBlock *mb, Class *exception, unsigne
                 if(!isInstanceOf(caught_class, exception))
                     continue;
             }
-            return mb->code + table[i].handler_pc;
+            return ((CodePntr)mb->code) + table[i].handler_pc;
         }
 
     return NULL;
 }
     
-unsigned char *findCatchBlock(Class *exception) {
+CodePntr findCatchBlock(Class *exception) {
     Frame *frame = getExecEnv()->last_frame;
-    unsigned char *handler_pc = NULL;
+    CodePntr handler_pc = NULL;
 
     while(((handler_pc = findCatchBlockInMethod(frame->mb, exception, frame->last_pc)) == NULL)
-		    && (frame->prev->mb != NULL)) {
+                    && (frame->prev->mb != NULL)) {
 
         if(frame->mb->access_flags & ACC_SYNCHRONIZED) {
             Object *sync_ob = frame->mb->access_flags & ACC_STATIC ?
-		    (Object*)frame->mb->class : (Object*)frame->lvars[0];
+                    (Object*)frame->mb->class : (Object*)frame->lvars[0];
             objectUnlock(sync_ob);
         }
         frame = frame->prev;
@@ -161,8 +169,8 @@ unsigned char *findCatchBlock(Class *exception) {
     return handler_pc;
 }
 
-int mapPC2LineNo(MethodBlock *mb, unsigned char *pc_pntr) {
-    int pc = pc_pntr - mb->code;
+int mapPC2LineNo(MethodBlock *mb, CodePntr pc_pntr) {
+    int pc = pc_pntr - (CodePntr) mb->code;
     int i;
 
     if(mb->line_no_table_size > 0) {
@@ -232,23 +240,23 @@ Object *convertStackTrace(Object *vmthrwble) {
 
     for(i = 0, j = 0; i < depth; j++) {
         MethodBlock *mb = (MethodBlock*)src[i++];
-	unsigned char *pc = (unsigned char *)src[i++];
-	ClassBlock *cb = CLASS_CB(mb->class);
-	unsigned char *dot_name = slash2dots(cb->name);
+        CodePntr pc = (CodePntr)src[i++];
+        ClassBlock *cb = CLASS_CB(mb->class);
+        unsigned char *dot_name = slash2dots(cb->name);
 
-	int isNative = mb->access_flags & ACC_NATIVE ? TRUE : FALSE;
+        int isNative = mb->access_flags & ACC_NATIVE ? TRUE : FALSE;
         Object *filename = isNative ? NULL : (cb->source_file_name ?
                                              createString(cb->source_file_name) : NULL);
-	Object *classname = createString(dot_name);
-	Object *methodname = createString(mb->name);
-	Object *ste = allocObject(ste_class);
+        Object *classname = createString(dot_name);
+        Object *methodname = createString(mb->name);
+        Object *ste = allocObject(ste_class);
         free(dot_name);
 
         if(exceptionOccured())
             return NULL;
 
-	executeMethod(ste, vmthrow_init_mb, filename, isNative ? -1 : mapPC2LineNo(mb, pc),
-			classname, methodname, isNative);
+        executeMethod(ste, vmthrow_init_mb, filename, isNative ? -1 : mapPC2LineNo(mb, pc),
+                        classname, methodname, isNative);
 
         if(exceptionOccured())
             return NULL;
