@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -26,7 +26,7 @@
 #include "thread.h"
 #include "lock.h"
 
-#define JNI_VERSION JNI_VERSION_1_2
+#define JNI_VERSION JNI_VERSION_1_4
 
 /* The extra used in expanding the local refs table.
  * This must be >= size of JNIFrame to be thread safe
@@ -46,8 +46,41 @@ void Jam_DeleteGlobalRef(JNIEnv *env, jobject obj);
 JNIFrame *ensureJNILrefCapacity(int cap);
 static void initJNIGrefs();
 
+/* Cached values initialised on startup for JNI 1.4 NIO support */
+static int buffCap_offset, buffAddr_offset, rawdata_offset;
+static Class *buffImpl_class, *rawdata_class;
+static MethodBlock *buffImpl_init_mb;
+static char nio_init_OK = FALSE;
+
 void initialiseJNI() {
+    FieldBlock *buffCap_fb, *buffAddr_fb, *rawdata_fb;
+    Class *buffer_class;
+
+    /* Initialise the global references table */
     initJNIGrefs();
+
+    /* Cache class and method/fields for JNI 1.4 NIO support */
+
+    buffer_class = findSystemClass0("java/nio/Buffer");
+    buffImpl_class = findSystemClass0("java/nio/DirectByteBufferImpl");
+    rawdata_class = findSystemClass0("gnu/classpath/RawData32");
+
+    buffImpl_init_mb = findMethod(buffImpl_class, "<init>",
+                      "(Ljava/lang/Object;Lgnu/classpath/RawData;III)V");
+
+    buffCap_fb = findField(buffer_class, "cap", "I");
+    rawdata_fb = findField(rawdata_class, "data", "I");
+    buffAddr_fb = findField(buffer_class, "address", "Lgnu/classpath/RawData;");
+
+    if(buffImpl_init_mb == NULL || buffCap_fb == NULL ||
+             rawdata_fb == NULL || buffAddr_fb == NULL) {
+        return;
+    }
+
+    buffCap_offset = buffCap_fb->offset;
+    buffAddr_offset = buffAddr_fb->offset;
+    rawdata_offset = rawdata_fb->offset;
+    nio_init_OK = TRUE;
 }
 
 int initJNILrefs() {
@@ -79,7 +112,7 @@ JNIFrame *ensureJNILrefCapacity(int cap) {
 
     if(size < cap) {
         int incr = cap-size;
-	if(incr < sizeof(JNIFrame)/sizeof(Object*))
+        if(incr < sizeof(JNIFrame)/sizeof(Object*))
             incr = sizeof(JNIFrame)/sizeof(Object*);
 
         if((frame = expandJNILrefs(ee, frame, incr)) == NULL)
@@ -113,8 +146,8 @@ void delJNILref(Object *ref) {
     for(; opntr < frame->next_ref; opntr++)
         if(*opntr == ref) {
             *opntr = NULL;
-	    return;
-	}
+            return;
+        }
 }
  
 JNIFrame *pushJNILrefFrame(int cap) {
@@ -165,23 +198,23 @@ static Object *addJNIGref(Object *ref) {
         int new_size;
         Object **new_table;
         int i, j;
-	       
-	if(global_ref_deleted >= LIST_INCR) {
+               
+        if(global_ref_deleted >= LIST_INCR) {
             new_size = global_ref_size;
-	    new_table = global_ref_table;
-	} else {
-	    new_size = global_ref_size + LIST_INCR - global_ref_deleted;
-       	    new_table = (Object**)sysMalloc(new_size * sizeof(Object*));
-	}
+            new_table = global_ref_table;
+        } else {
+            new_size = global_ref_size + LIST_INCR - global_ref_deleted;
+            new_table = (Object**)sysMalloc(new_size * sizeof(Object*));
+        }
 
         for(i = 0, j = 0; i < global_ref_size; i++)
             if(global_ref_table[i])
                 new_table[j++] = global_ref_table[i];
 
-	global_ref_next = j;
-	global_ref_size = new_size;
-	global_ref_table = new_table;
-	global_ref_deleted = 0;
+        global_ref_next = j;
+        global_ref_size = new_size;
+        global_ref_table = new_table;
+        global_ref_deleted = 0;
     }
 
     global_ref_table[global_ref_next++] = ref;
@@ -202,9 +235,9 @@ static void delJNIGref(Object *ref) {
     for(i = 0; i < global_ref_next; i++)
         if(global_ref_table[i] == ref) {
             global_ref_table[i] = NULL;
-	    global_ref_deleted++;
-	    break;
-	}
+            global_ref_deleted++;
+            break;
+        }
 
     unlockVMLock(global_ref_lock, self);
     enableSuspend(self);
@@ -225,6 +258,55 @@ void markJNIGlobalRefs() {
     enableSuspend(self);
 }
 
+/* Extensions added to JNI in JDK 1.4 */
+
+jobject Jam_NewDirectByteBuffer(JNIEnv *env, void *addr, jlong capacity) {
+    Object *buff, *rawdata;
+
+    if(!nio_init_OK)
+        return NULL;
+
+    if((buff = allocObject(buffImpl_class)) != NULL &&
+            (rawdata = allocObject(rawdata_class)) != NULL) {
+
+        INST_DATA(rawdata)[rawdata_offset] = (u4)addr;
+        executeMethod(buff, buffImpl_init_mb, NULL, rawdata, (int)capacity,
+                      (int)capacity, 0);
+    }
+
+    return (jobject) addJNILref(buff);
+}
+
+static void *Jam_GetDirectBufferAddress(JNIEnv *env, jobject buffer) {
+    Object *buff = (Object*)buffer;
+
+    if(!nio_init_OK)
+        return NULL;
+
+    if(buff != NULL) {
+        Object *rawdata = (Object *)INST_DATA(buff)[buffAddr_offset];
+        if(rawdata != NULL)
+            return (void*)INST_DATA(rawdata)[rawdata_offset];
+    }
+
+    return NULL;
+}
+
+jlong Jam_GetDirectBufferCapacity(JNIEnv *env, jobject buffer) {
+    Object *buff = (Object*)buffer;
+
+    if(!nio_init_OK)
+        return -1;
+
+    if(buff != NULL) {
+        Object *rawdata = (Object *)INST_DATA(buff)[buffAddr_offset];
+        if(rawdata != NULL)
+            return (jlong)INST_DATA(buff)[buffCap_offset];
+    }
+
+    return -1;
+}
+
 /* Extensions added to JNI in JDK 1.2 */
 
 jmethodID Jam_FromReflectedMethod(JNIEnv *env, jobject method) {
@@ -238,13 +320,13 @@ jfieldID Jam_FromReflectedField(JNIEnv *env, jobject field) {
 jobject Jam_ToReflectedMethod(JNIEnv *env, jclass cls, jmethodID methodID, jboolean isStatic) {
     MethodBlock *mb = (MethodBlock *)methodID;
     if(strcmp(mb->name, "<init>") == 0)
-        return (jobject) createConstructorObject(mb);
+        return (jobject) createReflectConstructorObject(mb);
     else
-        return (jobject) createMethodObject(mb);
+        return (jobject) createReflectMethodObject(mb);
 }
 
 jobject Jam_ToReflectedField(JNIEnv *env, jclass cls, jfieldID fieldID, jboolean isStatic) {
-    return (jobject) createFieldObject((FieldBlock *)fieldID);
+    return (jobject) createReflectFieldObject((FieldBlock *)fieldID);
 }
 
 jint Jam_PushLocalFrame(JNIEnv *env, jint capacity) {
@@ -516,14 +598,14 @@ jarray Jam_NewObjectArray(JNIEnv *env, jsize length, jclass elementClass, jobjec
     array_class = findArrayClassFromClass(ac_name, (Class*)elementClass);
     if(array_class != NULL) {
         Object *array = allocArray(array_class, length, 4);
-	if(array != NULL) {
+        if(array != NULL) {
             if(initialElement != NULL) {
                 u4 *data = INST_DATA(array) + 1;
 
-	        while(length--)
+                while(length--)
                    *data++ = (u4) initialElement;
             }
-	    return (jarray) addJNILref(array);
+            return (jarray) addJNILref(array);
         }
     }
     return NULL;
@@ -585,7 +667,7 @@ native_type Jam_GetStatic##type##Field(JNIEnv *env, jclass clazz, jfieldID field
 
 #define SET_STATIC_FIELD(type, native_type)                                                \
 void Jam_SetStatic##type##Field(JNIEnv *env, jclass clazz, jfieldID fieldID,               \
-		native_type value) {                                                       \
+                native_type value) {                                                       \
     FieldBlock *fb = (FieldBlock *) fieldID;                                               \
     *(native_type *)&fb->static_value = value;                                             \
 }
@@ -660,7 +742,7 @@ native_type Jam_Call##type##MethodA(JNIEnv *env, jobject obj, jmethodID mID, jva
 
 #define NONVIRTUAL_METHOD(type, native_type)                                                     \
 native_type Jam_CallNonvirtual##type##Method(JNIEnv *env, jobject obj, jclass clazz,             \
-		jmethodID methodID, ...) {                                                       \
+                jmethodID methodID, ...) {                                                       \
     native_type *ret;                                                                            \
     va_list jargs;                                                                               \
                                                                                                  \
@@ -673,20 +755,20 @@ native_type Jam_CallNonvirtual##type##Method(JNIEnv *env, jobject obj, jclass cl
 }                                                                                                \
                                                                                                  \
 native_type Jam_CallNonvirtual##type##MethodV(JNIEnv *env, jobject obj, jclass clazz,            \
-		jmethodID methodID, va_list jargs) {                                             \
+                jmethodID methodID, va_list jargs) {                                             \
     return *(native_type*)                                                                       \
-	      executeMethodVaList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, jargs);   \
+              executeMethodVaList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, jargs);   \
 }                                                                                                \
                                                                                                  \
 native_type Jam_CallNonvirtual##type##MethodA(JNIEnv *env, jobject obj, jclass clazz,            \
-		jmethodID methodID, jvalue *jargs) {                                             \
+                jmethodID methodID, jvalue *jargs) {                                             \
     return *(native_type*)                                                                       \
-	    executeMethodList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs);  \
+            executeMethodList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs);  \
 }
 
 #define STATIC_METHOD(type, native_type)                                                         \
 native_type Jam_CallStatic##type##Method(JNIEnv *env, jclass clazz,                              \
-		jmethodID methodID, ...) {                                                       \
+                jmethodID methodID, ...) {                                                       \
     native_type *ret;                                                                            \
     va_list jargs;                                                                               \
                                                                                                  \
@@ -698,15 +780,15 @@ native_type Jam_CallStatic##type##Method(JNIEnv *env, jclass clazz,
 }                                                                                                \
                                                                                                  \
 native_type Jam_CallStatic##type##MethodV(JNIEnv *env, jclass clazz, jmethodID methodID,         \
-		va_list jargs) {                                                                 \
+                va_list jargs) {                                                                 \
     return *(native_type*)                                                                       \
-	    executeMethodVaList(NULL, (Class*)clazz, (MethodBlock*)methodID, jargs);             \
+            executeMethodVaList(NULL, (Class*)clazz, (MethodBlock*)methodID, jargs);             \
 }                                                                                                \
                                                                                                  \
 native_type Jam_CallStatic##type##MethodA(JNIEnv *env, jclass clazz, jmethodID methodID,         \
-		jvalue *jargs) {                                                                 \
+                jvalue *jargs) {                                                                 \
     return *(native_type*)                                                                       \
-	    executeMethodList(NULL, (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs);          \
+            executeMethodList(NULL, (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs);          \
 }
 
 #define CALL_METHOD(access)         \
@@ -754,26 +836,26 @@ jobject Jam_CallObjectMethodA(JNIEnv *env, jobject obj, jmethodID methodID, jval
 }
 
 jobject Jam_CallNonvirtualObjectMethod(JNIEnv *env, jobject obj, jclass clazz,
-		jmethodID methodID, ...) {
+                jmethodID methodID, ...) {
     Object *ret;
     va_list jargs;
     va_start(jargs, methodID);
     ret = addJNILref(*(Object**) executeMethodVaList((Object*)obj,
-			    (Class*)clazz, (MethodBlock*)methodID, jargs));
+                            (Class*)clazz, (MethodBlock*)methodID, jargs));
     va_end(jargs);
     return (jobject)ret;
 }
 
 jobject Jam_CallNonvirtualObjectMethodV(JNIEnv *env, jobject obj, jclass clazz,
-		jmethodID methodID, va_list jargs) {
+                jmethodID methodID, va_list jargs) {
     return (jobject)addJNILref(*(Object**) executeMethodVaList((Object*)obj,
-			    (Class*)clazz, (MethodBlock*)methodID, jargs));
+                            (Class*)clazz, (MethodBlock*)methodID, jargs));
 }
 
 jobject Jam_CallNonvirtualObjectMethodA(JNIEnv *env, jobject obj, jclass clazz,
-		jmethodID methodID, jvalue *jargs) {
+                jmethodID methodID, jvalue *jargs) {
     return (jobject)addJNILref(*(Object**) executeMethodList((Object*)obj,
-			    (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs));
+                            (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs));
 }
 
 jobject Jam_CallStaticObjectMethod(JNIEnv *env, jclass clazz, jmethodID methodID, ...) {
@@ -781,36 +863,46 @@ jobject Jam_CallStaticObjectMethod(JNIEnv *env, jclass clazz, jmethodID methodID
     va_list jargs;
     va_start(jargs, methodID);
     ret = addJNILref(*(Object**) executeMethodVaList(NULL,
-			      (Class*)clazz, (MethodBlock*)methodID, jargs));
+                              (Class*)clazz, (MethodBlock*)methodID, jargs));
     va_end(jargs);
     return (jobject)ret;
 }
 
 jobject Jam_CallStaticObjectMethodV(JNIEnv *env, jclass clazz,
-		jmethodID methodID, va_list jargs) {
+                jmethodID methodID, va_list jargs) {
     return (jobject)addJNILref(*(Object**) executeMethodVaList(NULL,
-		              (Class*)clazz, (MethodBlock*)methodID, jargs));
+                              (Class*)clazz, (MethodBlock*)methodID, jargs));
 }
 
 jobject Jam_CallStaticObjectMethodA(JNIEnv *env, jclass clazz,
-		jmethodID methodID, jvalue *jargs) {
+                jmethodID methodID, jvalue *jargs) {
     return (jobject)addJNILref(*(Object**) executeMethodList(NULL,
-		            (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs));
+                            (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs));
 }
 
 void Jam_CallVoidMethod(JNIEnv *env, jobject obj, jmethodID methodID, ...) {
     va_list jargs;
+    MethodBlock *mb;
+    Object *ob = (Object *)obj;
+ 
     va_start(jargs, methodID);
-    executeMethodVaList((Object*)obj, ((Object*)obj)->class, (MethodBlock*)methodID, jargs);
+    if((mb = lookupVirtualMethod(ob, (MethodBlock*)methodID)) != NULL)
+        executeMethodVaList(ob, ob->class, mb, jargs);
     va_end(jargs);
 }
 
 void Jam_CallVoidMethodV(JNIEnv *env, jobject obj, jmethodID methodID, va_list jargs) {
-    executeMethodVaList((Object*)obj, ((Object*)obj)->class, (MethodBlock*)methodID, jargs);
+    MethodBlock *mb;
+    Object *ob = (Object *)obj;
+    if((mb = lookupVirtualMethod(ob, (MethodBlock*)methodID)) != NULL)
+        executeMethodVaList(ob, ob->class, mb, jargs);
 }
 
 void Jam_CallVoidMethodA(JNIEnv *env, jobject obj, jmethodID methodID, jvalue *jargs) {
-    executeMethodList((Object*)obj, ((Object*)obj)->class, (MethodBlock*)methodID, (u8*)jargs);
+    MethodBlock *mb;
+    Object *ob = (Object *)obj;
+    if((mb = lookupVirtualMethod(ob, (MethodBlock*)methodID)) != NULL)
+        executeMethodList(ob, ob->class, mb, (u8*)jargs);
 }
 
 void Jam_CallNonvirtualVoidMethod(JNIEnv *env, jobject obj, jclass clazz, jmethodID methodID, ...) {
@@ -821,12 +913,12 @@ void Jam_CallNonvirtualVoidMethod(JNIEnv *env, jobject obj, jclass clazz, jmetho
 }
 
 void Jam_CallNonvirtualVoidMethodV(JNIEnv *env, jobject obj, jclass clazz,
-		jmethodID methodID, va_list jargs) {
+                jmethodID methodID, va_list jargs) {
       executeMethodVaList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, jargs);
 }
 
 void Jam_CallNonvirtualVoidMethodA(JNIEnv *env, jobject obj, jclass clazz,
-		jmethodID methodID, jvalue *jargs) {
+                jmethodID methodID, jvalue *jargs) {
     executeMethodList((Object*)obj, (Class*)clazz, (MethodBlock*)methodID, (u8*)jargs);
 }
 
@@ -910,32 +1002,32 @@ PRIM_ARRAY_OP(Double, jdouble, T_DOUBLE);
     Jam_##direction##type##field_type##Field
 
 #define FIELDS2(direction, type)              \
-	FIELD(direction, type,  Object),      \
-	FIELD(direction, type, Boolean),      \
-	FIELD(direction, type,  Byte),        \
-	FIELD(direction, type, Char),         \
-	FIELD(direction, type, Short),        \
-	FIELD(direction, type, Int),          \
-	FIELD(direction, type, Long),         \
-	FIELD(direction, type, Float),        \
-	FIELD(direction, type, Double)
+        FIELD(direction, type,  Object),      \
+        FIELD(direction, type, Boolean),      \
+        FIELD(direction, type,  Byte),        \
+        FIELD(direction, type, Char),         \
+        FIELD(direction, type, Short),        \
+        FIELD(direction, type, Int),          \
+        FIELD(direction, type, Long),         \
+        FIELD(direction, type, Float),        \
+        FIELD(direction, type, Double)
 
 #define FIELDS(type)                          \
-	FIELDS2(Get, type),                   \
-	FIELDS2(Set, type)
+        FIELDS2(Get, type),                   \
+        FIELDS2(Set, type)
 
 #define ARRAY(op, el_type, type)              \
-	Jam_##op##el_type##Array##type
+        Jam_##op##el_type##Array##type
 
 #define ARRAY_OPS(op, type)                   \
-	ARRAY(op, Boolean, type),             \
-	ARRAY(op, Byte, type),                \
-	ARRAY(op, Char, type),                \
-	ARRAY(op, Short, type),               \
-	ARRAY(op, Int, type),                 \
-	ARRAY(op, Long, type),                \
-	ARRAY(op, Float, type),               \
-	ARRAY(op, Double, type)
+        ARRAY(op, Boolean, type),             \
+        ARRAY(op, Byte, type),                \
+        ARRAY(op, Char, type),                \
+        ARRAY(op, Short, type),               \
+        ARRAY(op, Int, type),                 \
+        ARRAY(op, Long, type),                \
+        ARRAY(op, Float, type),               \
+        ARRAY(op, Double, type)
 
 struct _JNINativeInterface Jam_JNINativeInterface = {
     NULL,
@@ -1010,7 +1102,10 @@ struct _JNINativeInterface Jam_JNINativeInterface = {
     Jam_ReleaseStringCritical,
     Jam_NewWeakGlobalRef,
     Jam_DeleteWeakGlobalRef,
-    Jam_ExceptionCheck
+    Jam_ExceptionCheck,
+    Jam_NewDirectByteBuffer,
+    Jam_GetDirectBufferAddress,
+    Jam_GetDirectBufferCapacity
 };
 
 jint Jam_DestroyJavaVM(JavaVM *vm) {
@@ -1023,6 +1118,11 @@ jint Jam_AttachCurrentThread(JavaVM *vm, void **penv, void *args) {
     exitVM(1);
 }
 
+jint Jam_AttachCurrentThreadAsDaemon(JavaVM *vm, void **penv, void *args) {
+    fprintf(stderr, "JNI - FatalError: AttachCurrentThreadAsDaemon unimplemented.\n");
+    exitVM(1);
+}
+
 jint Jam_DetachCurrentThread(JavaVM *vm) {
     fprintf(stderr, "JNI - FatalError: DetachCurrentThread unimplemented.\n");
     exitVM(1);
@@ -1032,7 +1132,8 @@ struct _JNINativeInterface Jam_JNINativeInterface;
 static void *env = &Jam_JNINativeInterface;
 
 jint Jam_GetEnv(JavaVM *vm, void **penv, jint version) {
-    if((version != JNI_VERSION_1_2) && (version != JNI_VERSION_1_1)) {
+    if((version != JNI_VERSION_1_4) && (version != JNI_VERSION_1_2)
+                                    && (version != JNI_VERSION_1_1)) {
         *penv = NULL;
         return JNI_EVERSION;
     }
@@ -1053,7 +1154,8 @@ struct _JNIInvokeInterface Jam_JNIInvokeInterface = {
     Jam_DestroyJavaVM,
     Jam_AttachCurrentThread,
     Jam_DetachCurrentThread,
-    Jam_GetEnv
+    Jam_GetEnv,
+    Jam_AttachCurrentThreadAsDaemon,
 };
 
 jint JNI_GetCreatedJavaVMs(JavaVM **buff, jsize buff_len, jsize *num) {
@@ -1064,4 +1166,8 @@ jint JNI_GetCreatedJavaVMs(JavaVM **buff, jsize buff_len, jsize *num) {
     }
     return JNI_ERR;
 }
+
+void *getJNIInterface() {
+    return &Jam_JNINativeInterface;
+}
 #endif
