@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -28,12 +28,10 @@
 #include <string.h>
 #include <errno.h>
 
-/* For system properties */
-#include <sys/utsname.h>
-
 #include "jam.h"
 #include "thread.h"
 #include "lock.h"
+#include "natives.h"
 
 static int pd_offset;
 
@@ -95,9 +93,8 @@ u4 *arraycopy(Class *class, MethodBlock *mb, u4 *ostack) {
     int start2 = ostack[3];
     int length = ostack[4];
 
-
     if((src == NULL) || (dest == NULL))
-	signalException("java/lang/NullPointerException", NULL);
+        signalException("java/lang/NullPointerException", NULL);
     else {
         ClassBlock *scb = CLASS_CB(src->class);
         ClassBlock *dcb = CLASS_CB(dest->class);
@@ -111,13 +108,14 @@ u4 *arraycopy(Class *class, MethodBlock *mb, u4 *ostack) {
                         || ((start1 + length) > sdata[0]) || ((start2 + length) > ddata[0])) {
             signalException("java/lang/ArrayIndexOutOfBoundsException", NULL);
             return ostack;
-	}
+        }
 
         if(isInstanceOf(dest->class, src->class)) {
             int size;
 
             switch(scb->name[1]) {
                 case 'B':
+                case 'Z':
                     size = 1;
                     break;
                 case 'C':
@@ -139,31 +137,31 @@ u4 *arraycopy(Class *class, MethodBlock *mb, u4 *ostack) {
             memmove(((char *)&ddata[1]) + start2*size,
                     ((char *)&sdata[1]) + start1*size,
                     length*size);
-	} else {
-	    Object **sob, **dob;
-	    int i;
+        } else {
+            Object **sob, **dob;
+            int i;
 
-   	    if(!(((scb->name[1] == 'L') || (scb->name[1] == '[')) &&
-   	                  ((dcb->name[1] == 'L') || (dcb->name[1] == '['))))
+            if(!(((scb->name[1] == 'L') || (scb->name[1] == '[')) &&
+                          ((dcb->name[1] == 'L') || (dcb->name[1] == '['))))
                 goto storeExcep; 
 
-	    /* Not compatible array types, but elements may be compatible...
-	       e.g. src = [Ljava/lang/Object, dest = [Ljava/lang/String, but
+            /* Not compatible array types, but elements may be compatible...
+               e.g. src = [Ljava/lang/Object, dest = [Ljava/lang/String, but
                all src = Strings - check one by one...
-	     */
-	    
-	    if(scb->dim > dcb->dim)
+             */
+            
+            if(scb->dim > dcb->dim)
                 goto storeExcep;
 
-	    sob = (Object**)&sdata[start1+1];
-	    dob = (Object**)&ddata[start2+1];
+            sob = (Object**)&sdata[start1+1];
+            dob = (Object**)&ddata[start2+1];
 
-	    for(i = 0; i < length; i++) {
+            for(i = 0; i < length; i++) {
                 if((*sob != NULL) && !arrayStoreCheck(dest->class, (*sob)->class))
                     goto storeExcep;
                 *dob++ = *sob++;
             }
-	}
+        }
     }
     return ostack;
 
@@ -227,66 +225,15 @@ u4 *nativeGetLibname(Class *class, MethodBlock *mb, u4 *ostack) {
     return ostack;
 }
 
-void setProperty(Object *this, char *key, char *value) {
-    Object *k = Cstr2String(key);
-    Object *v = Cstr2String(value ? value : "?");
-
-    MethodBlock *mb = lookupMethod(this->class, "put",
-                           "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
-    executeMethod(this, mb, k, v);
+u4 *propertiesPreInit(Class *class, MethodBlock *mb, u4 *ostack) {
+    Object *properties = (Object *)*ostack;
+    addDefaultProperties(properties);
+    return ostack;
 }
 
-extern Property *commandline_props;
-extern int commandline_props_count;
-
-u4 *insertSystemProperties(Class *class, MethodBlock *mb, u4 *ostack) {
-    Object *this = (Object *)*ostack;
-    struct utsname info;
-
-    uname(&info);
-
-    setProperty(this, "java.vm.name", "JamVM");
-    setProperty(this, "java.vm.version", VERSION);
-    setProperty(this, "java.vm.vendor", "Robert Lougher");
-    setProperty(this, "java.vm.vendor.url", "http://jamvm.sourceforge.net");
-    setProperty(this, "java.version", "1.4");
-    setProperty(this, "java.vendor", "GNU Classpath");
-    setProperty(this, "java.vendor.url", "http://gnu.classpath.org");
-    setProperty(this, "java.home", INSTALL_DIR);
-    setProperty(this, "java.specification.version", "1.4");
-    setProperty(this, "java.specification.vendor", "Sun Microsystems, Inc.");
-    setProperty(this, "java.specification.name", "Java Platform API Specification");
-    setProperty(this, "java.vm.specification.version", "1.0");
-    setProperty(this, "java.vm.specification.vendor", "Sun Microsystems, Inc.");
-    setProperty(this, "java.vm.specification.name", "Java Virtual Machine Specification");
-    setProperty(this, "java.class.version", "48.0");
-    setProperty(this, "java.class.path", getClassPath());
-    setProperty(this, "java.boot.class.path", getBootClassPath());
-    setProperty(this, "java.library.path", getDllPath());
-    setProperty(this, "java.io.tmpdir", "/tmp");
-    setProperty(this, "java.compiler", "");
-    setProperty(this, "java.ext.dirs", "");
-    setProperty(this, "os.name", info.sysname);
-    setProperty(this, "os.arch", info.machine);
-    setProperty(this, "os.version", info.release);
-    setProperty(this, "file.separator", "/");
-    setProperty(this, "path.separator", ":");
-    setProperty(this, "line.separator", "\n");
-    setProperty(this, "user.name", getenv("USER"));
-    setProperty(this, "user.home", getenv("HOME"));
-    setProperty(this, "user.dir", getenv("PWD"));
-
-    /* Handle command line properties */
-
-    if(commandline_props_count) {
-        int i;
-        for(i = 0; i < commandline_props_count; i++) {
-            setProperty(this, commandline_props[i].key, commandline_props[i].value);
-            free(commandline_props[i].key);
-        }
-        free(commandline_props);
-    }
-
+u4 *propertiesPostInit(Class *class, MethodBlock *mb, u4 *ostack) {
+    Object *properties = (Object *)*ostack;
+    addCommandLineProperties(properties);
     return ostack;
 }
 
@@ -425,7 +372,7 @@ u4 *forName0(u4 *ostack, int resolve, Object *loader) {
     char *cstr;
     
     if(string == NULL) {
-	signalException("java/lang/NullPointerException", NULL);
+        signalException("java/lang/NullPointerException", NULL);
         return ostack;
     }
 
@@ -462,8 +409,9 @@ u4 *forName0(u4 *ostack, int resolve, Object *loader) {
 
 out:
     if(class == NULL) {
+        Object *e = exceptionOccured();
         clearException();
-	signalException("java/lang/ClassNotFoundException", cstr);
+        signalChainedException("java/lang/ClassNotFoundException", cstr, e);
     } else
         if(resolve)
             initClass(class);
@@ -514,44 +462,85 @@ u4 *getStackTrace(Class *class, MethodBlock *m, u4 *ostack) {
     return ostack;
 }
 
-/* java.lang.VMSecurityManager */
+/* gnu.classpath.VMStackWalker */
 
-u4 *currentClassLoader(Class *class, MethodBlock *mb, u4 *ostack) {
-    *ostack++ = (u4) getCurrentClassLoader();
+Frame *skipFrames(Frame *last) {
+
+loop:
+    /* Skip the frame with the found class, and
+       check if the previous is a dummy frame */
+    if((last = last->prev)->mb == NULL) {
+
+        /* Skip the dummy frame, and check if
+         * we're at the top of the stack */
+        if((last = last->prev)->prev == NULL)
+            return NULL;
+
+        /* Check if we were invoked via reflection */
+        if(last->mb->class == getReflectMethodClass()) {
+
+            /* There will be two frames for invoke.  Skip
+               the first, and jump back.  This also handles
+               recursive invocation via reflection. */
+
+            last = last->prev;
+            goto loop;
+        }
+    }
+    return last;
+}
+
+Class *getCallingClass0() {
+    Frame *last = getExecEnv()->last_frame->prev;
+
+    if((last->mb == NULL && (last = last->prev)->prev == NULL) ||
+             (last = skipFrames(last)) == NULL)
+        return NULL;
+
+    return last->mb->class;
+}
+
+u4 *getCallingClass(Class *class, MethodBlock *mb, u4 *ostack) {
+    *ostack++ = (int) getCallingClass0();
     return ostack;
 }
+
+u4 *getCallingClassLoader(Class *clazz, MethodBlock *mb, u4 *ostack) {
+    Class *class = getCallingClass0();
+
+    *ostack++ = (int) (class ? CLASS_CB(class)->class_loader : NULL);
+    return ostack;
+}
+
 u4 *getClassContext(Class *class, MethodBlock *mb, u4 *ostack) {
     Class *class_class = findArrayClass("[Ljava/lang/Class;");
-    Frame *bottom, *last;
-    int depth = 0;
     Object *array;
-    int *data;
+    Frame *last;
 
     if(class_class == NULL)
         return ostack;
 
-    /* The top frame will be frame for VMSecurityManager - skip */
-    last = getExecEnv()->last_frame->prev;
+    if((last = skipFrames(getExecEnv()->last_frame)) == NULL)
+        array = allocArray(class_class, 0, 4); 
+    else {
+        Frame *bottom = last;
+        int depth = 0;
 
-    /* The next frame _may_ be SecurityManager */
-    if(last->mb && strcmp(CLASS_CB(last->mb->class)->name, "java/lang/SecurityManager") == 0)
-        last = last->prev;
-
-    bottom = last;
-    do {
-        for(; last->mb != NULL; last = last->prev, depth++);
-    } while((last = last->prev)->prev != NULL);
-    
-    array = allocArray(class_class, depth, 4);
-
-    if(array != NULL) {
-        data = INST_DATA(array);
-
-        depth = 1;
         do {
-            for(; bottom->mb != NULL; bottom = bottom->prev)
-                data[depth++] = (int)bottom->mb->class;
-        } while((bottom = bottom->prev)->prev != NULL);
+            for(; last->mb != NULL; last = last->prev, depth++);
+        } while((last = last->prev)->prev != NULL);
+    
+        array = allocArray(class_class, depth, 4);
+
+        if(array != NULL) {
+            int *data = INST_DATA(array);
+
+            depth = 1;
+            do {
+                for(; bottom->mb != NULL; bottom = bottom->prev)
+                    data[depth++] = (int)bottom->mb->class;
+            } while((bottom = bottom->prev)->prev != NULL);
+        }
     }
 
     *ostack++ = (int)array;
@@ -607,7 +596,7 @@ u4 *findLoadedClass(Class *clazz, MethodBlock *mb, u4 *ostack) {
     int len, i;
 
     if(string == NULL) {
-	signalException("java/lang/NullPointerException", NULL);
+        signalException("java/lang/NullPointerException", NULL);
         return ostack;
     }
 
@@ -678,7 +667,8 @@ Object *getAndCheckObject(u4 *ostack, Class *type) {
         return NULL;
     }
     if(!isInstanceOf(type, ob->class)) {
-        signalException("java/lang/IllegalArgumentException", "object is not an instance of declaring class");
+        signalException("java/lang/IllegalArgumentException",
+                        "object is not an instance of declaring class");
         return NULL;
     }
     return ob;
@@ -864,88 +854,218 @@ u4 *holdsLock(Class *class, MethodBlock *mb, u4 *ostack) {
     return ostack;
 }
 
-char *native_methods[][2] = {
-                             "arraycopy",		(char*)arraycopy,
-                             "insertSystemProperties",	(char*)insertSystemProperties,
-                             "getPrimitiveClass",	(char*)getPrimitiveClass,
-			     "defineClass",             (char*)defineClass0,
-			     "findLoadedClass",         (char*)findLoadedClass,
-			     "resolveClass",            (char*)resolveClass0,
-                             "intern",			(char*)intern,
-                             "loadClass",		(char*)loadClass,
-                             "loadArrayClass",		(char*)loadArrayClass,
-                             "initialize",		(char*)initialize,
-                             "throwException",		(char*)throwException,
-                             "hasClassInitializer",	(char*)hasClassInitializer,
-                             "forName",			(char*)forName,
-			     "isInstance",		(char*)isInstance,
-			     "isAssignableFrom",	(char*)isAssignableFrom,
-			     "isInterface",		(char*)isInterface,
-			     "isPrimitive",		(char*)isPrimitive,
-			     "isArray",			(char*)isArray,
-                             "getSuperclass",		(char*)getSuperclass,
-                             "getName",			(char*)getName,
-                             "getClass",		(char*)getClass,
-			     "getComponentType",	(char*)getComponentType,
-                             "identityHashCode",	(char*)identityHashCode,
-			     "clone",			(char*)jamClone,
-			     "wait",			(char*)wait,
-			     "notify",			(char*)notify,
-			     "notifyAll",		(char*)notifyAll,
-                             "gc",			(char*)gc,
-                             "runFinalization",		(char*)runFinalization,
-                             "exit",			(char*)exitInternal,
-                             "fillInStackTrace",	(char*)fillInStackTrace,
-                             "getStackTrace",		(char*)getStackTrace,
-			     "currentClassLoader",	(char*)currentClassLoader,
-			     "getClassContext",		(char*)getClassContext,
-			     "getClassLoader",		(char*)getClassLoader,
-			     "constructNative",		(char*)constructNative,
-			     "invokeNative",		(char*)invokeNative,
-                             "nativeLoad",		(char*)nativeLoad,
-                             "nativeGetLibname",	(char*)nativeGetLibname,
-			     "freeMemory",		(char*)freeMemory,
-			     "totalMemory",		(char*)totalMemory,
-			     "maxMemory",		(char*)maxMemory,
-			     "currentThread",		(char*)currentThread,
-			     "create",			(char*)create,
-			     "sleep",			(char*)jamSleep,
-			     "interrupt",		(char*)interrupt,
-			     "isAlive",			(char*)isAlive,
-			     "yield",			(char*)yield,
-			     "isInterrupted",		(char*)isInterrupted,
-			     "interrupted",		(char*)interrupted,
-			     "nativeSetPriority",	(char*)nativeSetPriority,
-                             "holdsLock",		(char*)holdsLock,
-			     "getDeclaredClasses",	(char*)getDeclaredClasses,
-			     "getDeclaringClass",	(char*)getDeclaringClass0,
-			     "getDeclaredConstructors",	(char*)getDeclaredConstructors,
-			     "getDeclaredMethods",	(char*)getDeclaredMethods,
-			     "getDeclaredFields",	(char*)getDeclaredFields,
-			     "getInterfaces",		(char*)getInterfaces,
-			     "getModifiers",		(char*)getClassModifiers,
-			     "getConstructorModifiers",	(char*)getMethodModifiers,
-			     "getMethodModifiers",	(char*)getMethodModifiers,
-			     "getFieldModifiers",	(char*)getFieldModifiers,
-                             "getField",		(char*)getField,
-                             "setField",		(char*)setField,
-                             "setZField",		(char*)setPrimitiveField,
-                             "setBField",		(char*)setPrimitiveField,
-                             "setCField",		(char*)setPrimitiveField,
-                             "setSField",		(char*)setPrimitiveField,
-                             "setIField",		(char*)setPrimitiveField,
-                             "setFField",		(char*)setPrimitiveField,
-                             "setJField",		(char*)setPrimitiveField,
-                             "setDField",		(char*)setPrimitiveField,
-                             "getZField",		(char*)getPrimitiveField,
-                             "getBField",		(char*)getPrimitiveField,
-                             "getCField",		(char*)getPrimitiveField,
-                             "getSField",		(char*)getPrimitiveField,
-                             "getIField",		(char*)getPrimitiveField,
-                             "getFField",		(char*)getPrimitiveField,
-                             "getJField",		(char*)getPrimitiveField,
-                             "getDField",		(char*)getPrimitiveField,
-                             "getBootClassPathSize",    (char*)getBootClassPathSize,
-                             "getBootClassPathResource",(char*)getBootClassPathResource,
-                             NULL,			NULL};
+/* java.security.VMAccessController */
 
+/* instance method getStack()[[Ljava/lang/Object; */
+u4 *getStack(Class *class, MethodBlock *mb, u4 *ostack) {
+    Class *object_class = findArrayClass("[[Ljava/lang/Object;");
+    Class *class_class = findArrayClass("[Ljava/lang/Class;");
+    Class *string_class = findArrayClass("[Ljava/lang/String;");
+    Object *stack, *names, *classes;
+    Frame *frame;
+    int depth = 0;
+
+    if(object_class == NULL || class_class == NULL || string_class == NULL)
+      return ostack;
+
+    frame = getExecEnv()->last_frame;
+
+    do {
+        for(; frame->mb != NULL; frame = frame->prev, depth++);
+    } while((frame = frame->prev)->prev != NULL);
+
+    stack = allocArray(object_class, 2, 4);
+    classes = allocArray(class_class, depth, 4);
+    names = allocArray(string_class, depth, 4);
+
+    if(stack != NULL && names != NULL && classes != NULL) {
+        int *dcl = INST_DATA(classes);
+        int *dnm = INST_DATA(names);
+
+        frame = getExecEnv()->last_frame;
+        depth = 1;
+
+        do {
+            for(; frame->mb != NULL; frame = frame->prev, depth++) {
+                dcl[depth] = (int) frame->mb->class;
+                dnm[depth] = (int) createString(frame->mb->name);
+            }
+        } while((frame = frame->prev)->prev != NULL);
+
+        INST_DATA(stack)[1] = (int) classes;
+        INST_DATA(stack)[2] = (int) names;
+    }
+
+    *ostack++ = (int) stack;
+    return ostack;
+}
+
+u4 *currentClassLoader(Class *class, MethodBlock *mb, u4 *ostack) {
+    *ostack++ = (u4) getCurrentClassLoader();
+    return ostack;
+}
+
+VMMethod vm_object[] = {
+    {"getClass",                    getClass},
+    {"clone",                       jamClone},
+    {"wait",                        wait},
+    {"notify",                      notify},
+    {"notifyAll",                   notifyAll},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_system[] = {
+    {"arraycopy",                   arraycopy},
+    {"identityHashCode",            identityHashCode},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_runtime[] = {
+    {"freeMemory",                  freeMemory},
+    {"totalMemory",                 totalMemory},
+    {"maxMemory",                   maxMemory},
+    {"gc",                          gc},
+    {"runFinalization",             runFinalization},
+    {"exit",                        exitInternal},
+    {"nativeLoad",                  nativeLoad},
+    {"nativeGetLibname",            nativeGetLibname},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_class[] = {
+    {"isInstance",                  isInstance},
+    {"isAssignableFrom",            isAssignableFrom},
+    {"isInterface",                 isInterface},
+    {"isPrimitive",                 isPrimitive},
+    {"isArray",                     isArray},
+    {"getSuperclass",               getSuperclass},
+    {"getComponentType",            getComponentType},
+    {"getName",                     getName},
+    {"getDeclaredClasses",          getDeclaredClasses},
+    {"getDeclaringClass",           getDeclaringClass0},
+    {"getDeclaredConstructors",     getDeclaredConstructors},
+    {"getDeclaredMethods",          getDeclaredMethods},
+    {"getDeclaredFields",           getDeclaredFields},
+    {"getInterfaces",               getInterfaces},
+    {"getClassLoader",              getClassLoader},
+    {"getModifiers",                getClassModifiers},
+    {"forName",                     forName},
+    {"loadArrayClass",              loadArrayClass},
+    {"initialize",                  initialize},
+    {"throwException",              throwException},
+    {"hasClassInitializer",         hasClassInitializer},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_string[] = {
+    {"intern",                      intern},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_thread[] = {
+    {"currentThread",               currentThread},
+    {"create",                      create},
+    {"sleep",                       jamSleep},
+    {"interrupt",                   interrupt},
+    {"isAlive",                     isAlive},
+    {"yield",                       yield},
+    {"isInterrupted",               isInterrupted},
+    {"interrupted",                 interrupted},
+    {"nativeSetPriority",           nativeSetPriority},
+    {"holdsLock",                   holdsLock},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_throwable[] = {
+    {"fillInStackTrace",            fillInStackTrace},
+    {"getStackTrace",               getStackTrace},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_classloader[] = {
+    {"loadClass",                   loadClass},
+    {"getPrimitiveClass",           getPrimitiveClass},
+    {"defineClass",                 defineClass0},
+    {"findLoadedClass",             findLoadedClass},
+    {"resolveClass",                resolveClass0},
+    {"getBootClassPathSize",        getBootClassPathSize},
+    {"getBootClassPathResource",    getBootClassPathResource},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_reflect_constructor[] = {
+    {"constructNative",             constructNative},
+    {"getConstructorModifiers",     getMethodModifiers},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_reflect_method[] = {
+    {"invokeNative",                invokeNative},
+    {"getMethodModifiers",          getMethodModifiers},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_reflect_field[] = {
+    {"getFieldModifiers",           getFieldModifiers},
+    {"getField",                    getField},
+    {"setField",                    setField},
+    {"setZField",                   setPrimitiveField},
+    {"setBField",                   setPrimitiveField},
+    {"setCField",                   setPrimitiveField},
+    {"setSField",                   setPrimitiveField},
+    {"setIField",                   setPrimitiveField},
+    {"setFField",                   setPrimitiveField},
+    {"setJField",                   setPrimitiveField},
+    {"setDField",                   setPrimitiveField},
+    {"getZField",                   getPrimitiveField},
+    {"getBField",                   getPrimitiveField},
+    {"getCField",                   getPrimitiveField},
+    {"getSField",                   getPrimitiveField},
+    {"getIField",                   getPrimitiveField},
+    {"getFField",                   getPrimitiveField},
+    {"getJField",                   getPrimitiveField},
+    {"getDField",                   getPrimitiveField},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_system_properties[] = {
+    {"preInit",                     propertiesPreInit},
+    {"postInit",                    propertiesPostInit},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_stack_walker[] = {
+    {"getClassContext",             getClassContext},
+    {"getCallingClass",             getCallingClass},
+    {"getCallingClassLoader",       getCallingClassLoader},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_access_controller[] = {
+    {"getStack",                    getStack},
+    {NULL,                          NULL}
+};
+
+VMMethod vm_object_input_stream[] = {
+    {"currentClassLoader",          currentClassLoader},
+    {NULL,                          NULL}
+};
+
+VMClass native_methods[] = {
+    {"java/lang/VMClass",                vm_class},
+    {"java/lang/VMObject",               vm_object},
+    {"java/lang/VMThread",               vm_thread},
+    {"java/lang/VMSystem",               vm_system},
+    {"java/lang/VMString",               vm_string},
+    {"java/lang/VMRuntime",              vm_runtime},
+    {"java/lang/VMThrowable",            vm_throwable},
+    {"java/lang/VMClassLoader",          vm_classloader},
+    {"java/lang/reflect/Field",          vm_reflect_field},
+    {"java/lang/reflect/Method",         vm_reflect_method},
+    {"java/lang/reflect/Constructor",    vm_reflect_constructor},
+    {"java/security/VMAccessController", vm_access_controller},
+    {"gnu/classpath/VMSystemProperties", vm_system_properties},
+    {"java/io/ObjectInputStream",        vm_object_input_stream},
+    {"gnu/classpath/VMStackWalker",      vm_stack_walker},
+    {NULL,                               NULL}
+};
