@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -550,6 +550,29 @@ Object *invoke(Object *ob, MethodBlock *mb, Object *arg_array, Object *param_typ
     return ret;
 }
 
+/* Reflection access from JNI */
+
+Object *createReflectConstructorObject(MethodBlock *mb) {
+    if(!inited && !initReflection())
+        return NULL;
+
+    return createConstructorObject(mb);
+}
+
+Object *createReflectMethodObject(MethodBlock *mb) {
+    if(!inited && !initReflection())
+        return NULL;
+
+    return createMethodObject(mb);
+}
+
+Object *createReflectFieldObject(FieldBlock *fb) {
+    if(!inited && !initReflection())
+        return NULL;
+
+    return createFieldObject(fb);
+}
+
 MethodBlock *mbFromReflectObject(Object *reflect_ob) {
     if(reflect_ob->class == cons_reflect_class)
         return (MethodBlock*)INST_DATA(reflect_ob)[cons_slot_offset];
@@ -561,3 +584,8 @@ FieldBlock *fbFromReflectObject(Object *reflect_ob) {
     return (FieldBlock*)INST_DATA(reflect_ob)[field_slot_offset];
 }
 
+/* Needed for stack walking */
+
+Class *getReflectMethodClass() {
+    return method_reflect_class;
+}
