@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004 Robert Lougher <rob@lougher.demon.co.uk>.
+ * Copyright (C) 2003, 2004, 2005 Robert Lougher <rob@lougher.demon.co.uk>.
  *
  * This file is part of JamVM.
  *
@@ -21,6 +21,7 @@
 #include <stdio.h>
 #include <string.h>
 #include "jam.h"
+#include "arch.h"
 
 MethodBlock *findMethod(Class *class, char *methodname, char *type) {
    ClassBlock *cb = CLASS_CB(class);
@@ -47,7 +48,7 @@ FieldBlock *findField(Class *class, char *fieldname, char *type) {
            if(strcmp(fb->type, type) == 0)
                return fb;
            else
-	       return NULL;
+               return NULL;
         }
 
     return NULL;
@@ -94,18 +95,18 @@ Class *resolveClass(Class *class, int cp_index, int init) {
 
 retry:
     switch(CP_TYPE(cp, cp_index)) {
-	case CONSTANT_Locked:
-	    goto retry;
+        case CONSTANT_Locked:
+            goto retry;
 
         case CONSTANT_Resolved:
             resolved_class = (Class *)CP_INFO(cp, cp_index);
-	    break;
+            break;
 
-	case CONSTANT_Class: {
-	    char *classname;
+        case CONSTANT_Class: {
+            char *classname;
             int name_idx = CP_CLASS(cp, cp_index);
 
-	    if(CP_TYPE(cp, cp_index) != CONSTANT_Class)
+            if(CP_TYPE(cp, cp_index) != CONSTANT_Class)
                 goto retry;
 
             classname = CP_UTF8(cp, name_idx);
@@ -118,10 +119,12 @@ retry:
                 return NULL;
 
             CP_TYPE(cp, cp_index) = CONSTANT_Locked;
+            MBARRIER();
             CP_INFO(cp, cp_index) = (u4)resolved_class;
+            MBARRIER();
             CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
 
-	    break;
+            break;
         }
     }
 
@@ -137,21 +140,21 @@ MethodBlock *resolveMethod(Class *class, int cp_index) {
 
 retry:
     switch(CP_TYPE(cp, cp_index)) {
-	case CONSTANT_Locked:
-	    goto retry;
+        case CONSTANT_Locked:
+            goto retry;
 
         case CONSTANT_Resolved:
             mb = (MethodBlock *)CP_INFO(cp, cp_index);
-	    break;
+            break;
 
         case CONSTANT_Methodref: {
-	    Class *resolved_class;
+            Class *resolved_class;
             ClassBlock *resolved_cb;
             char *methodname, *methodtype;
             int cl_idx = CP_METHOD_CLASS(cp, cp_index);
             int name_type_idx = CP_METHOD_NAME_TYPE(cp, cp_index);
 
-	    if(CP_TYPE(cp, cp_index) != CONSTANT_Methodref)
+            if(CP_TYPE(cp, cp_index) != CONSTANT_Methodref)
                 goto retry;
 
             methodname = CP_UTF8(cp, CP_NAME_TYPE_NAME(cp, name_type_idx));
@@ -179,12 +182,14 @@ retry:
                 initClass(mb->class);
 
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
+                MBARRIER();
                 CP_INFO(cp, cp_index) = (u4)mb;
+                MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
                 signalException("java/lang/NoSuchMethodError", methodname);
 
-	    break;
+            break;
         }
     }
 
@@ -197,20 +202,20 @@ MethodBlock *resolveInterfaceMethod(Class *class, int cp_index) {
 
 retry:
     switch(CP_TYPE(cp, cp_index)) {
-	case CONSTANT_Locked:
-	    goto retry;
+        case CONSTANT_Locked:
+            goto retry;
 
         case CONSTANT_Resolved:
             mb = (MethodBlock *)CP_INFO(cp, cp_index);
-	    break;
+            break;
 
         case CONSTANT_InterfaceMethodref: {
-	    Class *resolved_class;
+            Class *resolved_class;
             char *methodname, *methodtype;
             int cl_idx = CP_METHOD_CLASS(cp, cp_index);
             int name_type_idx = CP_METHOD_NAME_TYPE(cp, cp_index);
 
-	    if(CP_TYPE(cp, cp_index) != CONSTANT_InterfaceMethodref)
+            if(CP_TYPE(cp, cp_index) != CONSTANT_InterfaceMethodref)
                 goto retry;
 
             methodname = CP_UTF8(cp, CP_NAME_TYPE_NAME(cp, name_type_idx));
@@ -230,19 +235,21 @@ retry:
                 int i;
 
                 for(i = 0; mb == NULL && (i < cb->imethod_table_size); i++) {
-	            Class *intf = cb->imethod_table[i].interface;
+                    Class *intf = cb->imethod_table[i].interface;
                     mb = findMethod(intf, methodname, methodtype);
                 }
             }
 
             if(mb) {
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
+                MBARRIER();
                 CP_INFO(cp, cp_index) = (u4)mb;
+                MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
                 signalException("java/lang/NoSuchMethodError", methodname);
 
-	    break;
+            break;
         }
     }
 
@@ -255,20 +262,20 @@ FieldBlock *resolveField(Class *class, int cp_index) {
 
 retry:
     switch(CP_TYPE(cp, cp_index)) {
-	case CONSTANT_Locked:
-	    goto retry;
+        case CONSTANT_Locked:
+            goto retry;
 
         case CONSTANT_Resolved:
             fb = (FieldBlock *)CP_INFO(cp, cp_index);
-	    break;
+            break;
 
         case CONSTANT_Fieldref: {
-	    Class *resolved_class;
+            Class *resolved_class;
             char *fieldname, *fieldtype;
             int cl_idx = CP_FIELD_CLASS(cp, cp_index);
             int name_type_idx = CP_FIELD_NAME_TYPE(cp, cp_index);
 
-	    if(CP_TYPE(cp, cp_index) != CONSTANT_Fieldref)
+            if(CP_TYPE(cp, cp_index) != CONSTANT_Fieldref)
                 goto retry;
 
             fieldname = CP_UTF8(cp, CP_NAME_TYPE_NAME(cp, name_type_idx));
@@ -284,12 +291,14 @@ retry:
                 initClass(fb->class);
 
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
+                MBARRIER();
                 CP_INFO(cp, cp_index) = (u4)fb;
+                MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             } else
                 signalException("java/lang/NoSuchFieldError", fieldname);
 
-	    break;
+            break;
         }
     }
  
@@ -301,12 +310,17 @@ u4 resolveSingleConstant(Class *class, int cp_index) {
 
 retry:
     switch(CP_TYPE(cp, cp_index)) {
-	case CONSTANT_Locked:
-	    goto retry;
+        case CONSTANT_Locked:
+            goto retry;
 
-	case CONSTANT_String: {
+        case CONSTANT_Class:
+            resolveClass(class, cp_index, FALSE);
+            break;
+
+        case CONSTANT_String: {
             Object *string;
             int idx = CP_STRING(cp, cp_index);
+
             if(CP_TYPE(cp, cp_index) != CONSTANT_String)
                 goto retry;
 
@@ -314,7 +328,9 @@ retry:
 
             if(string) {
                 CP_TYPE(cp, cp_index) = CONSTANT_Locked;
+                MBARRIER();
                 CP_INFO(cp, cp_index) = (u4)findInternedString(string);
+                MBARRIER();
                 CP_TYPE(cp, cp_index) = CONSTANT_Resolved;
             }
 
@@ -338,9 +354,12 @@ MethodBlock *lookupVirtualMethod(Object *ob, MethodBlock *mb) {
     if(CLASS_CB(mb->class)->access_flags & ACC_INTERFACE) {
         int i;
 
-        for(i = 0; (i < cb->imethod_table_size) && (mb->class != cb->imethod_table[i].interface); i++);
+        for(i = 0; (i < cb->imethod_table_size) &&
+                   (mb->class != cb->imethod_table[i].interface); i++);
+
         if(i == cb->imethod_table_size) {
-            signalException("java/lang/IncompatibleClassChangeError", "unimplemented interface");
+            signalException("java/lang/IncompatibleClassChangeError",
+                            "unimplemented interface");
             return NULL;
         }
         mtbl_idx = cb->imethod_table[i].offsets[mtbl_idx];
@@ -355,3 +374,38 @@ MethodBlock *lookupVirtualMethod(Object *ob, MethodBlock *mb) {
 
     return mb;
 }
+
+/* This function is used when rewriting a field access bytecode
+   in the direct-threaded interpreter.  We need to know how many
+   slots are used on the stack, but the field reference may not
+   be resolved, and resolving at preparation will break Java's
+   lazy resolution semantics. */
+
+#ifdef DIRECT
+int peekIsFieldLong(Class *class, int cp_index) {
+    ConstantPool *cp = &(CLASS_CB(class)->constant_pool);
+    char *type;
+
+retry:
+    switch(CP_TYPE(cp, cp_index)) {
+        case CONSTANT_Locked:
+            goto retry;
+
+        case CONSTANT_Resolved:
+            type = ((FieldBlock *)CP_INFO(cp, cp_index))->type;
+            break;
+
+        case CONSTANT_Fieldref: {
+            int name_type_idx = CP_FIELD_NAME_TYPE(cp, cp_index);
+
+            if(CP_TYPE(cp, cp_index) != CONSTANT_Fieldref)
+                goto retry;
+
+            type = CP_UTF8(cp, CP_NAME_TYPE_TYPE(cp, name_type_idx));
+            break;
+        }
+    }
+ 
+    return *type == 'J' || *type == 'D';
+}
+#endif
