@@ -29,7 +29,7 @@
 #include "lock.h"
 
 #ifdef TRACETHREAD
-#define TRACE(x) printf x
+#define TRACE(x) printf x; fflush(stdout)
 #else
 #define TRACE(x)
 #endif
@@ -105,7 +105,7 @@ retry:
             int n = ffs(~tidBitmap[i]);
             tidBitmap[i] |= 1 << (n-1);
             return (i<<5) + n;
-	}
+        }
     }
 
     tidBitmap = (unsigned int *)sysRealloc(tidBitmap,
@@ -154,9 +154,9 @@ void threadInterrupt(Thread *thread) {
         if(!owner) {
             int i;
 
-	    /* Another thread may be holding the monitor -
-	     * if we can't get it after a couple attempts give-up
-	     * to avoid deadlock */
+            /* Another thread may be holding the monitor -
+             * if we can't get it after a couple attempts give-up
+             * to avoid deadlock */
             for(i = 0; i < INTERRUPT_TRIES; i++) {
                 if(!pthread_mutex_trylock(&mon->lock))
                     goto got_lock;
@@ -261,18 +261,18 @@ void *threadStart(void *arg) {
     group = (Object *)INST_DATA(jThread)[group_offset];
     if((excep = exceptionOccured())) {
         Class *throwable;
-	MethodBlock *uncaught_exp;
+        MethodBlock *uncaught_exp;
        
-	clearException();
-	throwable = findSystemClass0("java/lang/Throwable");
+        clearException();
+        throwable = findSystemClass0("java/lang/Throwable");
         if(throwable && isInstanceOf(throwable, excep->class)
                      && (uncaught_exp = lookupMethod(group->class, "uncaughtException",
-			                              "(Ljava/lang/Thread;Ljava/lang/Throwable;)V")))
+                                                      "(Ljava/lang/Thread;Ljava/lang/Throwable;)V")))
             executeMethod(group, uncaught_exp, jThread, excep);
-	else {
+        else {
             setException(excep);
             printException();
-	}
+        }
     }
 
     /* remove thread from thread group */
@@ -308,7 +308,7 @@ void *threadStart(void *arg) {
 
     if(non_daemon_thrds == 0) {
         /* No need to bother with disabling suspension
-	 * around lock, as we're no longer on thread list */
+         * around lock, as we're no longer on thread list */
         pthread_mutex_lock(&exit_lock);
         pthread_cond_signal(&exit_cv);
         pthread_mutex_unlock(&exit_lock);
@@ -334,7 +334,7 @@ void createJavaThread(Object *jThread, long long stack_size) {
         pthread_mutex_unlock(&lock);
         enableSuspend(self);
         signalException("java/lang/IllegalThreadStateException", "thread already started");
-	return;
+        return;
     }
 
     ee = (ExecEnv*)sysMalloc(sizeof(ExecEnv));
@@ -345,6 +345,7 @@ void createJavaThread(Object *jThread, long long stack_size) {
     thread->ee = ee;
     ee->thread = jThread;
     ee->stack_size = stack_size;
+
     INST_DATA(vmthread)[vmData_offset] = (u4)thread;
     INST_DATA(vmthread)[thread_offset] = (u4)jThread;
     INST_DATA(jThread)[vmthread_offset] = (u4)vmthread;
@@ -353,10 +354,10 @@ void createJavaThread(Object *jThread, long long stack_size) {
     if(pthread_create(&thread->tid, &attributes, threadStart, thread)) {
         INST_DATA(jThread)[vmthread_offset] = 0;
         free(ee);
-	free(thread);
+        free(thread);
         enableSuspend(self);
-	signalException("java/lang/OutOfMemoryError", "can't create thread");
-	return;
+        signalException("java/lang/OutOfMemoryError", "can't create thread");
+        return;
     }
 
     pthread_mutex_lock(&lock);
@@ -390,18 +391,14 @@ void createJavaThread(Object *jThread, long long stack_size) {
     enableSuspend(self);
 }
 
-Thread *attachThread(char *name, char is_daemon, void *stack_base) {
+Thread *attachThread(char *name, char is_daemon, void *stack_base, Thread *thread) {
     ExecEnv *ee;
-    Thread *thread;
     Object *vmthread;
 
     ee = (ExecEnv*)sysMalloc(sizeof(ExecEnv));
-    thread = (Thread*)sysMalloc(sizeof(Thread));
     memset(ee, 0, sizeof(ExecEnv));
-    memset(thread, 0, sizeof(Thread));
 
     thread->tid = pthread_self();
-    thread->state = RUNNING;
     thread->stack_base = stack_base;
     thread->ee = ee;
 
@@ -432,6 +429,9 @@ Thread *attachThread(char *name, char is_daemon, void *stack_base) {
         non_daemon_thrds++;
 
     thread->id = genThreadID();
+    thread->state = RUNNING;
+
+    pthread_cond_broadcast(&cv);
     pthread_mutex_unlock(&lock);
 
     TRACE(("Thread 0x%x id: %d attached\n", thread, thread->id));
@@ -440,12 +440,12 @@ Thread *attachThread(char *name, char is_daemon, void *stack_base) {
 
 static void *shell(void *args) {
     void *start = ((void**)args)[1];
-    Thread *self;
+    Thread *self = ((Thread**)args)[2];
 
     if(main_exited)
         return NULL;
 
-    self = attachThread(((char**)args)[0], TRUE, &self);
+    attachThread(((char**)args)[0], TRUE, &self, self);
 
     free(args);
     (*(void(*)(Thread*))start)(self);
@@ -453,12 +453,21 @@ static void *shell(void *args) {
 }
 
 void createVMThread(char *name, void (*start)(Thread*)) {
-    void **args = sysMalloc(2 * sizeof(void*));
+    Thread *thread = (Thread*)sysMalloc(sizeof(Thread));
+    void **args = sysMalloc(3 * sizeof(void*));
     pthread_t tid;
 
     args[0] = name;
     args[1] = start;
+    args[2] = thread;
+
+    memset(thread, 0, sizeof(Thread));
     pthread_create(&tid, &attributes, shell, args);
+
+    pthread_mutex_lock(&lock);
+    while(thread->state == 0)
+        pthread_cond_wait(&cv, &lock);
+    pthread_mutex_unlock(&lock);
 }
 
 void suspendAllThreads(Thread *self) {
@@ -470,15 +479,15 @@ void suspendAllThreads(Thread *self) {
     for(thread = &main_thread; thread != NULL; thread = thread->next) {
         if(thread == self)
             continue;
-	thread->suspend = TRUE;
-	if(!thread->blocking)
-	    pthread_kill(thread->tid, SIGUSR1);
+        thread->suspend = TRUE;
+        if(!thread->blocking)
+            pthread_kill(thread->tid, SIGUSR1);
     }
 
     for(thread = &main_thread; thread != NULL; thread = thread->next) {
         if(thread == self)
             continue;
-	while(!thread->blocking && thread->state != SUSPENDED)
+        while(!thread->blocking && thread->state != SUSPENDED)
             sched_yield();
     }
 
@@ -495,13 +504,13 @@ void resumeAllThreads(Thread *self) {
     for(thread = &main_thread; thread != NULL; thread = thread->next) {
         if(thread == self)
             continue;
-	thread->suspend = FALSE;
-	if(!thread->blocking)
-	    pthread_kill(thread->tid, SIGUSR1);
+        thread->suspend = FALSE;
+        if(!thread->blocking)
+            pthread_kill(thread->tid, SIGUSR1);
     }
 
     for(thread = &main_thread; thread != NULL; thread = thread->next) {
-	while(thread->state == SUSPENDED)
+        while(thread->state == SUSPENDED)
             sched_yield();
     }
 
@@ -572,19 +581,19 @@ void *dumpThreadsLoop(void *arg) {
     sigaddset(&mask, SIGINT);
 
     for(;;) {
-	sigwait(&mask, &sig);
+        sigwait(&mask, &sig);
 
-	if(sig == SIGINT)
+        if(sig == SIGINT)
             exitVM(0);
 
-	suspendAllThreads(&dummy);
+        suspendAllThreads(&dummy);
         printf("Thread Dump\n-----------\n\n");
         for(thread = &main_thread; thread != NULL; thread = thread->next) {
             char *name = String2Cstr((Object*)(INST_DATA(thread->ee->thread)[name_offset]));
             printf("Thread: %s 0x%x tid: %d state: %d\n", name, (int)thread, (int)thread->tid, thread->state);
-	    free(name);
+            free(name);
         }
-	resumeAllThreads(&dummy);
+        resumeAllThreads(&dummy);
     }
     return NULL;
 }
@@ -602,6 +611,7 @@ static void initialiseSignals() {
     sigemptyset(&mask);
     sigaddset(&mask, SIGQUIT);
     sigaddset(&mask, SIGINT);
+    sigaddset(&mask, SIGPIPE);
     sigprocmask(SIG_BLOCK, &mask, NULL);
 
     pthread_create(&tid, &attributes, dumpThreadsLoop, NULL);
@@ -730,7 +740,7 @@ void initialiseMainThread(int stack_size) {
     thrdGrp_class = findSystemClass("java/lang/ThreadGroup");
     if(exceptionOccured()) {
         printException();
-	exitVM(1);
+        exitVM(1);
     }
 
     root = findField(thrdGrp_class, "root", "Ljava/lang/ThreadGroup;");
@@ -752,4 +762,3 @@ error:
     fprintf(stderr, "Error initialising VM (initialiseMainThread)\n");
     exitVM(1);
 }
-
